/************ file.h **************
$Revision: 1.16 $
$Author: pb $
$Date: 2001/05/09 14:00:36 $
***********************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#ifndef _FILE_H_
#define _FILE_H_

/** The number of lines that are scanned for include statements 
    and MAKER_ADD_DEPENDENCY in the source files. */
#define MAXLINES 256
/** The maximum of dependencies a file can have. Both for header
    and for source files. */
#define MAXDEPEND 512
/** the maximum number of -Ixxx statements that can be collected
    for the compile command. */
#define MAXINCLUDEPATHS 2048

/** Number that signals that the file's extension is unknown. */
#define EXTUNKNOWN 0
/** Number that signals that the file is a header file. */
#define EXTH       1
/** Number that signals that the file is a source file. */
#define EXTCPP     2
/** Number that signals that the file is a object file. */
#define EXTOBJ     3

#include "makeutilities.h"

/** Remember to delete this! */
static class File *fileRoot=NULL;

/**
   Keeps all information, dependecies etc for one file.
   If getFileName returns NULL, then the file doesn't exist.

   Basically it is a linked list of File objects, but that
   is just a practical way of managing them.

   @author Philip Bergen
   @version 1.0
*/
class File
{
 public:
  /**
     Initializes all inner structures.
     @param fname The complete absolute or relative path to the file
     to represent.
  */
  File(const char *fname)
    {
      ageChecked=false;
      rebuild=false;
      scanned=false;
      compiled=false;
      dotted=false;
      subDotted=false;
      objected=false;
      included=false;
      tagged=false;
      hFileCount=0;
      cppFileCount=0;

      objFile=NULL;
      fullName=NULL;
      fileName=NULL;
      dotName=NULL;
      path=NULL;
      next=NULL;

      fullName=MakeUtilities::getStrCopy(fname);

      if (!getStats())
	{
	  extension=EXTUNKNOWN;
	  disectFilename();
	  if (extension==EXTUNKNOWN)
	    {
	      delete [] path;
	      path=NULL;
	      delete [] fileName;
	      fileName=NULL;
	    }
	  else
	    {
	      objFile=NULL;
	      for(int i=0;i<MAXDEPEND;i++)
		{
		  hFile[i]=NULL;
		  cppFile[i]=NULL;
		}
	      guess();
	    }
	}
    }

  /**
     Deletes paths and next.
  */
  ~File()
    {
      delete [] fullName;
      /* No need since it is part of path! */
      // delete [] fileName; 
      delete [] path;
      delete next;
    }

  /**
     @return The fileName without path. If NULL then the File does not
     exist.
  */
  const char *getFileName(void)
    {
      return fileName;
    }

  /**
     @return The path without fileName. If NULL then the File is in
     local directory.
  */
  const char *getPath(void)
    {
      return path;
    }

  /**
     @return The fullName, if NULL then the File is screwed.
  */
  const char *getFullName(void)
    {
      return fullName;
    }

  /**
     @return 0 for unknown file type, 1 for .h/.hh/.hpp files,
     2 for .c/.cc/.cpp files and 3 for .o/.obj files.
  */
  int getExtension(void)
    {
      return extension;
    }

  /**
     Adds the File object after itself, unless it is NULL in which case
     nothing happens. No sorting.
     @param file Any file object that is not already in the list!
     @return The supplied File object.
  */
  File *addNextFile(File *file)
    {
      if (file)
	{
	  if (next)
	    {
	      file->next=next;
	    }
	  next=file;
	}
      return file;
    }

  /**
     Parses the source-file (if its .c/.cc/.cpp or .h/.hh/.hpp) and
     extracts include statements. Standard path includes in brackets
     are ignored.

     All include statements after line 128 is ignored. First occurence 
     of the word 'class' on the beginning of a line terminates the 
     parsing of the file.     

     Recursively calls its dependencies and cppFile if any.

     File objects are sought through the fileRoot.
  */
  void resolveDependencies(void)
    {
      if (!scanned)
	{
	  if (extension==EXTH || extension==EXTCPP)
	    {
	      char buf[MAXLINES],*p,*q;
	      int line=0;
	      FILE *fin=fopen(fullName,"r");
	      if (fin)
		{
		  while (fgets(buf,MAXLINES,fin))
		    {
		      line++;
		      if (!strncmp(buf,"#include",8))
			{
			  p=strchr(buf+9,'\"');
			  if (p)
			    {
			      q=strchr(p+1,'\"');
			      if (q)
				{
				  *q=0;
				  addDependency(getFileNoPath(p+1));
				}
			    }
			  else
			    {
			      p=strchr(buf+9,'<');
			      if (p)
				{
				  q=strchr(p+1,'>');
				  if (q)
				    {			    
				      *q=0;
				      addDependency(getFileNoPath(p+1));
				    }
				}
			    }
			}
		      else if (!strncmp(buf,"MAKER_ADD_DEPENDENCY=",21))
			{
			  p=strchr(buf+21,'\"');
			  if (p)
			    {
			      q=strchr(p+1,'\"');
			      if (q)
				{
				  *q=0;
				  char *tmpPath=warpPath(p+1);
				  bool aCopy=false;
				  File *tmp=addDependency(findFile(tmpPath,
								   NULL,
								   aCopy));
				  if (!tmp)
				    {
				      printf("Dependency not found: %s\n",
					     tmpPath);
				    }
				  if (!aCopy)
				    {
				      addFile(tmp);
				    }
				}
			    }
			}
		      else
			{
			  if (!strncmp(buf,"class",5))
			    {
			      break;
			    }
			}
		      if (line>128)
			{
			  break;
			}
		    }
		  fclose(fin);
		}
	    }
	  scanned=true;
	  for(int i=0;i<cppFileCount;i++)
	    {
	      cppFile[i]->resolveDependencies();
	    }
	  for(int i=0;i<hFileCount;i++)
	    {
	      hFile[i]->resolveDependencies();
	    }
	}
    }

  /**
     Prints the contents of a file object.
     @param recursive If true then next is followed.
  */
  void print(bool recursive=false)
    {
      printf("This: 0x%X            %s\n",(int)this,rebuild?"REBUILD":"");
      printf("Fullname: %s\n",fullName);
      printf("Path and name: %s + %s\n",path,fileName);    
      printf("ObjFile: 0x%X\n",(int)objFile);
      printf("CppFiles:\t0x%X\n",(int)cppFile[0]);
      for(int i=1;i<cppFileCount;i++)
	{
	  printf("\t\t0x%X\n",(int)cppFile[i]);
	}
      printf("Next: 0x%X\n",(int)next);
      printf("Dependecies: %d\n",hFileCount);
      for(int i=0;i<hFileCount;i++)
	{
	  printf("%.7d: 0x%X\n",i,(int)hFile[i]);
	}
      if (recursive && next)
	{
	  next->print(recursive);
	}
    }

  /**
     Tags all files that need be rebuilt.
     Recursive.
  */
  void tagRebuild(void)
    {
      if (tagged)
	{
	  return;
	}
      tagged=true;
      if (extension==EXTCPP)
	{
	  rebuild=false;
	  if (!objFile)
	    {
	      rebuild=true;
	    }
	  else
	    {
	      clearAgeChecked();
	      if (isObjFileOld(objFile->stats.st_mtime) /*st_ctime?*/)
		{
		  rebuild=true;
		}
	    }
	}
      else if (extension==EXTH)
	{
	  for(int i=0;i<cppFileCount;i++)
	    {
	      cppFile[i]->tagRebuild();
	    }
	}
      for(int i=0;i<hFileCount;i++)
	{
	  hFile[i]->tagRebuild();
	}
    }

  /**
     Writes all compilation lines to a stream.
     Recursive.
     @param stream Any open stream.
     @param compiler The compiler string from OptionReader.
     @see OptionReader::getCompiler()
  */
  void writeCompilation(FILE *stream,const char *compiler)
    {
      if (compiled)
	{
	  return;
	}
      compiled=true;
      for(int i=0;i<cppFileCount;i++)
	{
	  cppFile[i]->writeCompilation(stream,compiler);
	}
      for(int i=0;i<hFileCount;i++)
	{
	  hFile[i]->writeCompilation(stream,compiler);
	}
      if (extension==EXTCPP && rebuild)
	{
	  char *output=MakeUtilities::changeExtension(fullName,"o");
	  fprintf(stream,"echo COMPILING: %s\n",fullName);
	  fprintf(stream,"%s -c -o %s",compiler,output);
	  delete [] output;

	  int count=0;
	  char *paths[MAXINCLUDEPATHS];
	  File::clearIncluded();
	  getIncludePath(paths,count);
	  for(int i=0;i<count;i++)
	    {
	      fprintf(stream," -I%s",paths[i]);
	    }
	  fprintf(stream," %s\n",fullName);
	  fprintf(stream,"if [ $? -ne 0 ]; then\nexit 1\nfi\n");
	}
    }

  /**
     Writes the final link directive to a stream.
     @param stream Any open stream.
     @param executable The target file name for the executable from
     OptionReader.
     @param linker The linker command from OptionReader.
     @param linkerTail The linker tailing string from OptionReader.
     @see OptionReader::getExecutable
     @see OptionReader::getLinker
     @see OptionReader::getLinkerTail
  */
  void writeLink(FILE* stream, const char *executable,
		 const char *linker, const char *linkerTail)
    {
      fprintf(stream,"echo LINKING: %s\n",executable);
      fprintf(stream,"%s -o %s",linker,executable);
      writeObjects(stream);
      fprintf(stream," %s\n",linkerTail);
    }

  /**
     Writes a number of dot-graphs.
     @param dotFile The dot file name from OptionReader.     
     @return True if successful.
     @see OptionReader::getDotoutput
  */
  bool writeDot(const char *dotFile)
  {
    char *fname=MakeUtilities::changeExtension(dotFile,
					       "all.dot");
    /* Build big messy graph. */
    FILE *fout=fopen(fname,"w+");
    if (fout)
      {
	fprintf(fout,"digraph \"All files\" {\n");
	fprintf(fout,"size=\"7.5,11\";ratio=fill;rankdir=LR;\n");
	fprintf(fout,"node [fontsize=11];\n");

	clearDotted();
	writeDotNodes(fout,true);
	fprintf(fout,"}\n");
	fclose(fout);
      }
    else
      {
	delete [] fname;
	return false;
      }
    delete [] fname;

    /* Build organized graph of CPPs only. */
    fout=fopen(dotFile,"w+");
    if (fout)
      {
	fprintf(fout,"digraph \"All source files\" {\n");
	fprintf(fout,"size=\"7.5,11\";ratio=fill;rankdir=LR;\n");
	fprintf(fout,"node [fontsize=11];\n");

	clearDotted();	
	clearSubDotted();
	writeDotCPPs(fout,NULL);

	fprintf(fout,"}\n");
	fclose(fout);
      }
    else
      {
	return false;
      }
    return true;
  }

private:
  /**
     @param incPath Any include path.
     @return A static string of a path less relative. Caller must NOT DELETE!
  */
  char *warpPath(char *incPath)
  {
    static char tmp[8000],*p;
    tmp[0]=0;
    
    if (!incPath || !path)
      {
	return incPath;
      }

    strcpy(tmp,path);
    char *addPath=incPath;
    while (!strncmp(addPath,"../",3))
      {	
	p=MakeUtilities::strchrReverse(tmp,'/');
	if (p)
	  {
	    *p=0;
	  }
	else 
	  {
	    if (!*tmp)
	      {
		printf("Cannot resolve relative path. Goes below /.\n");
		return incPath;
	      }
	    else
	      {
		*tmp=0;
	      }
	  }
	addPath=strchr(addPath,'/')+1;
      }
    if (tmp[strlen(tmp)]!='/')
      {
	strcat(tmp,"/");
      }
    strcat(tmp,addPath);
    return tmp;    
  }

  /**
     @param incPath Any include path.
     @return A pointer in the supplied string pointing at the filename 
     after the path. Caller must NOT DELETE!
  */
  static const char *stripPath(const char *incPath)
  {
    if (!incPath)
      {
	return incPath;
      }

    char *p=MakeUtilities::strchrReverse(incPath,'/');
    if (p)
      {
	return p+1;
      }
    return incPath;
  }

  /**
     Writes a dot graph of the dependencies.
     Recursive.
     @param stream Any open stream.
     @param followCPP If true the recursion goes into the CPP-files as well.
  */
  void writeDotNodes(FILE *stream, bool followCPP=false)
  {
    if (dotted)
      {
	return;
      }
    dotted=true;

    if (extension==EXTCPP)
      {
	if (rebuild)
	  {
	    fprintf(stream,"\"%s\" [label=\"%s\\n(rebuild)\", style=filled, "
		    "color=lightgray, shape=box];",fileName,fileName);
	  }
	else
	  {
	    fprintf(stream,"\"%s\" [shape=box];\n",fileName);
	  }
      }
    if (cppFileCount)
      {
	if (followCPP)
	  {
	    for(int i=0;i<cppFileCount;i++)
	      {
		cppFile[i]->writeDotNodes(stream,followCPP);
	      }
	  }
	for(int i=0;i<cppFileCount;i++)
	  {
	    fprintf(stream,"\"%s\" -> \"%s\";\n",
		    fileName,cppFile[i]->fileName);
	  }
      }
    if (hFileCount)
      {
	fprintf(stream,"\"%s\" -> { ",fileName);
	for (int i=0;i<hFileCount;i++)
	  {
	    fprintf(stream,"\"%s\"%c",hFile[i]->fileName,
		    (i+1)==hFileCount?' ':';');
	  }
	fprintf(stream,"}\n");
	for(int i=0;i<hFileCount;i++)
	  {
	    hFile[i]->writeDotNodes(stream,followCPP);
	  }
      }
  }

  /**
     Writes a dot graph of the CPP-files dependencies.
     Recursive.
     @param stream Any open stream.
     @param from The File that will be depending on this if this is a
     CPP-file.
  */
  void writeDotCPPs(FILE *stream, File *from)
  {
    if (extension==EXTCPP && from)
      {
	fprintf(stream,"\"%s\" -> \"%s\";\n",from->fileName,fileName);
      }
    if (subDotted)
      {
	return;
      }
    subDotted=true;
    if (extension==EXTCPP)
      {
	if (rebuild)
	  {
	    fprintf(stream,"\"%s\" [label=\"%s\\n(rebuild)\", style=filled, "
		    "color=lightgray, shape=box];",fileName,fileName);
	  }
	else
	  {
	    fprintf(stream,"\"%s\" [shape=box];\n",fileName);
	  }
	from=this;
      }
    for(int i=0;i<cppFileCount;i++)
      {
	cppFile[i]->writeDotCPPs(stream,from);
      }
    for(int i=0;i<hFileCount;i++)
      {
	hFile[i]->writeDotCPPs(stream,from);
      }
  }

  /**
     Determines the include path needed.
     Recursive.
     @param paths A list of previous paths (to avoid adding the same more
     than once).
     @param count The number of strings in paths. Is incremented when
     new paths are added.
  */
  void getIncludePath(char **paths, int &count)
    {
      if (included)
	{
	  return;
	}
      included=true;
      bool insert=true;
      for(int i=0;i<count;i++)
	{
	  if (!strcmp(path,paths[i]))
	    {
	      insert=false;
	      break;
	    }
	}
      if (insert)
	{
	  if (path && strlen(path))
	    {
	      paths[count]=MakeUtilities::getStrCopy(path);
	      count++;
	      if (count==MAXINCLUDEPATHS)
		{
		  printf("The include paths have overflown the buffer.\n");
		  _exit(-1);
		}
	    }
	}
      for(int i=0;i<hFileCount;i++)
	{
	  hFile[i]->getIncludePath(paths,count);
	}
    }

  /**
     Dumps all rebuilt objects (including path) to a stream.
     Recursive.
     @param stream Any stream that is open.
  */
  void writeObjects(FILE *stream)
    {
      if (objected)
	{
	  return;
	}
      objected=true;
      if (extension==EXTCPP)
	{
	  char *object=MakeUtilities::changeExtension(fullName,"o");
	  fprintf(stream," %s",object);
	  delete [] object;
	}
      for(int i=0;i<cppFileCount;i++)
	{
	  cppFile[i]->writeObjects(stream);
	}
      for(int i=0;i<hFileCount;i++)
	{
	  hFile[i]->writeObjects(stream);
	}
    }

  /**
     Asks recursively, following the header files from the dependencies,
     if any of them are newer than t.
     @param t The modification time of the object file.
     @return true if the objectFile is too old.
  */
  bool isObjFileOld(time_t t)
    {
      if (isNewer(t))
	{
	  return true;
	}
      if (ageChecked)
	{
	  return false;
	}
      ageChecked=true;
      for (int i=0;i<hFileCount;i++)
	{
	  if (hFile[i]->isObjFileOld(t))
	    {
	      return true;
	    }
	}
      return false;
    }

  /**
     @param t A UNIX-style time.
     @return true if the queried file is newer than t.
  */
  bool isNewer(time_t t)
    {
      return (t<stats.st_mtime);
    }

  /**
     Adds a File as a dependency. If MAXDEPEND is reached the program 
     is halted.
     @param file Any File object from a header file. If it is not
     a header file then nothing is added.
     @return The file-pointer.
  */
  File *addDependency(File *file)
    {
      if (file)
	{
	  if (file->extension==EXTH)
	    {
	      for (int i=0;i<hFileCount;i++)
		{
		  if (hFile[i]==file)
		    {
		      /* No need to add the same file twice. */
		      return file;
		    }
		}
	      hFile[hFileCount]=file;
	      hFileCount++;
	      if (hFileCount==MAXDEPEND)
		{
		  printf("Cannot insert dependency. Buffer overflow. "
			 "Probably files are including each other "
			 "recursively.\n");
		  _exit(-1);
		}
	    }
	  else if (file->extension==EXTCPP)
	    {
	      for (int i=0;i<cppFileCount;i++)
		{
		  if (cppFile[i]==file)
		    {
		      /* No need to add the same file twice. */
		      return file;
		    }
		}
	      cppFile[cppFileCount]=file;
	      cppFileCount++;
	      if (cppFileCount==MAXDEPEND)
		{
		  printf("Cannot insert dependency. Buffer overflow. "
			 "Probably files are including each other "
			 "recursively.\n");
		  _exit(-1);
		}
	    }
	  else if (file->extension==EXTOBJ)
	    {
	      if (objFile)
		{
		  printf("Cannot add more than one object file to a "
			 "source file.\n");
		  _exit(-1);
		}
	      objFile=file;
	    }
	}
      return file;
    }

  /**
     Gets stats about the File, if that isn't possible
     the return is non-zero.
  */
  int getStats(void)
    {
      return stat(fullName,&stats);
    }

  /**
     Splits the fullName into path and filename.
  */
  void disectFilename(void)
    {
      delete [] path;
      /* No need since it is part of path! */
      // delete [] fileName;
      path=MakeUtilities::getStrCopy(fullName);
      fileName=MakeUtilities::strchrReverse(path,'/');
      if (!fileName)
	{
	  fileName=path;
	  path=NULL;
	}
      else
	{
	  *fileName=0;
	  fileName++;
	}
      char *p=MakeUtilities::strchrReverse(fullName,'.');
      if (p)
	{
	  p++;
	  if (!strcmp(p,"h") || !strcmp(p,"hpp") || !strcmp(p,"hh"))
	    {
	      extension=EXTH;
	    }
	  else if (!strcmp(p,"c") || !strcmp(p,"cpp") || !strcmp(p,"cc"))
	    {
	      extension=EXTCPP;
	    }
	  else if (!strcmp(p,"o") || !strcmp(p,"obj"))
	    {
	      extension=EXTOBJ;
	    }

	  p=MakeUtilities::strchrReverse(fileName,'.');
	  if (p);
  	    {
	      *p=0;
	      dotName=MakeUtilities::getStrCopy(fileName,4);
	      *p='.';
	      strcat(dotName,"_");
	      strcat(dotName,p+1);
	    }
	}
    }

  /**
     If the extension is CPP then a File is added as objFile with a
     faked extension of obj.
  */
  void guess(void)
    {
      bool aCopy=false;
      File *tmp;
      if (extension==EXTH)
	{
	  tmp=addDependency(findFile(fullName,"cpp",aCopy));
	  if (!aCopy)
	    {
	      addFile(tmp);
	    }

	  tmp=addDependency(findFile(fullName,"cc",aCopy));
	  if (!aCopy)
	    {
	      addFile(tmp);
	    }

	  tmp=addDependency(findFile(fullName,"c",aCopy));
	  if (!aCopy)
	    {
	      addFile(tmp);
	    }
	}
      if (extension==EXTCPP)
	{
	  tmp=addDependency(findFile(fullName,"o",aCopy));
	  if (!aCopy)
	    {
	      addFile(tmp);
	    }

	  tmp=addDependency(findFile(fullName,"obj",aCopy));
	  if (!aCopy)
	    {
	      addFile(tmp);
	    }
	}
    }

  /**
     Sets all included flags to false to allow the building of a new
     include path.
     Recursive along next.
  */
  static void clearIncluded(void)
    {
      File *f=fileRoot;
      while (f)
	{
	  f->included=false;
	  f=f->next;
	}
    }

  /**
     Sets all ageChecked flags to false to allow a new object file
     to be checked for rebuild.
     Recursive along next.
  */
  static void clearAgeChecked(void)
    {
      File *f=fileRoot;
      while (f)
	{
	  f->ageChecked=false;
	  f=f->next;
	}
    }

  /**
     Sets all dotted flags to false to allow the building of new
     dot output.
     Recursive along next.
  */
  static void clearDotted(void)
    {
      File *f=fileRoot;
      while (f)
	{
	  f->dotted=false;
	  f=f->next;
	}
    }

  /**
     Sets all subDotted flags to false to allow the building of new
     dot output.
     Recursive along next.
  */
  static void clearSubDotted(void)
    {
      File *f=fileRoot;
      while (f)
	{
	  f->subDotted=false;
	  f=f->next;
	}
    }

  /**
     Attempts to create a File object from a filename and an extension
     (the current extension is switched to the supplied).
     @param fname Any filename with or without path.
     @param ext Any extension with out leading dot. If NULL the extension
     is left unchanged.
     @return A new File object if the file existed else NULL;
  */
  static File *findFile(const char *fname, const char *ext, bool &aCopy)
    {
      char *newName=MakeUtilities::getStrCopy(fname,ext?strlen(ext)+1:0);
      if (ext)
	{
	  char *p=MakeUtilities::strchrReverse(newName,'.');
	  if (!p)
	    {
	      p=newName+strlen(newName)+1;
	      *p='.';
	    }
	  p[1]=0;
	  strcat(newName,ext);
	}
      File *res=getFileNoPath(newName);
      if (res)
	{
	  delete [] newName;
	  aCopy=true;
	  return res;
	}
      aCopy=false;
      res=new File(newName);
      delete [] newName;
      if (!res->fileName)
	{
	  delete res;
	  res=NULL;
	}
      return res;
    }

  /**
     Runs through the linked list starting in fileRoot,
     searching for the file named fname. 
     
     <b>Comparing ONLY the fileName NOT the whole name and path!!!</B>
     
     @param fname Any file with .h/.hh/.hpp extension.
     @return A opinter to the File object or NULL if it is not in the chain.
  */
  static File *getFileNoPath(const char *fname)
    {
      File *f=fileRoot;
      while (f)
	{
	  if (!strcmp(f->fileName,stripPath(fname)))
	    {
	      return f;
	    }
	  f=f->next;
	}
      return NULL;
    }

public:
  /**
     Adds a file to the fileRoot.
     @param file Any file that has not already been added to the
     fileRoot.
  */
  static void addFile(File *file)
    {
      if (file)
	{
	  if (fileRoot)
	    {
	      fileRoot->addNextFile(file);
	    }
	  else
	    {
	      fileRoot=file;
	    }
	}
    }

  /**
     Prints the name of all files registered in the two pointers
     fileRoot and otherFilesRoot.
  */
  static void printRegistered(void)
    {
      File *f=fileRoot;
      while (f)
	{
	  f->print(false);
	  f=f->next;
	}
    }

private:
  char *fullName, *path, *fileName, *dotName;
  int extension;
  struct stat stats;
  File *objFile;
  File *cppFile[MAXDEPEND];
  File *hFile[MAXDEPEND];
  File *next;
  bool rebuild,scanned,tagged,compiled,objected,
    included,dotted,subDotted,ageChecked;
  int hFileCount,cppFileCount;
};

#endif

/*************** LOG *****************
 $Log: file.h,v $
 Revision 1.16  2001/05/09 14:00:36  pb
 Removed recusion bug in tagRebuild (isObjFileOld).

 Revision 1.15  2001/05/09 13:25:54  pb
 Fixed the bug caused in warpPath by the prior bugfix.

 Revision 1.14  2001/05/09 13:15:07  pb
 Fixed a bug in warpPath.

 Revision 1.13  2001/05/09 12:22:49  pb
 Reversed the compilation.

 Revision 1.12  2001/05/09 12:10:56  pb
 Cleaned up the source. Added new directive MAKER_ADD_DEPENDENCY.
 Added documentation.

 Revision 1.11  2001/05/08 18:10:18  pb
 Fixed a couple of bugs.

 Revision 1.10  2001/05/08 14:21:28  pb
 Works 100% and lots of new features.

 Revision 1.9  2001/05/07 17:42:00  pb
 Added conditional break if compile errors occur.

 Revision 1.8  2001/05/07 17:17:15  pb
 It really works!

 Revision 1.7  2001/05/01 07:33:25  pb
 Nothing really.

 Revision 1.6  2001/04/30 13:20:42  arnum
 Misc. small changes to unitests and xatof handles errors and matches strings
	
 Revision 1.5  2001/04/28 14:35:10  pb
 Added dot-output.
		
 Revision 1.4  2001/04/28 10:11:19  pb
 A -perhaps- working version.
		
 Revision 1.3  2001/04/27 14:23:36  pb
 It seems to start working. But there is a bug in building dependencies.

 Revision 1.2  2001/04/27 13:49:28  pb
 Added headers and footers.
*************************************/
