/************ file.h **************
$Revision: 1.10 $
$Author: pb $
$Date: 2001/05/09 14:52:14 $
***********************************/

#ifndef _OPTIONREADER_H_
#define _OPTIONREADER_H_

#include <time.h>
#include <unistd.h>
#include "makeutilities.h"

/**
   Reads an optionfile including, or command line parameters being, 
   any of these:
<PRE>
      MAINFILE=somefile.cpp
      EXECUTABLE=somefile
</PRE><I>Default: MAINFILE without extension</I>
<PRE>
      LINKER=g++ -lm -lxx -lxx -lxx
</PRE><I>Default: g++ -lm</I>
<PRE>
      LINKER_TAIL=libRPClort.a
</PRE><I>Default: nothing...</I>
<PRE>
      COMPILER=g++ -O2 -g -I/usr/local/lib/mysql
</PRE><I>Default: g++ -O2 -g</I>
<PRE>
      MAKESCRIPT=somefile.sh
</PRE><I>Default: MAINFILE with .sh-extension</I>
<PRE>
      DOTOUTPUT=somefile.dot
</PRE><I>Default: nothing...</I>
<PRE>
      MAKER_ADD_DEPENDENCY="somefile.xxx"
</PRE><I>Default: nothing...<br>
Note: The double quotes (") must be used. This option works only 
in source files.</I>
<PRE>
      NOFILE as parameter means no makfile is in parameter line.
</PRE>
<PRE>
      NOLINK as parameter means linking stage will be added to the
      makescript.
</PRE>
  @author Philip Bergen
  @version 1.0
*/
class OptionReader
{
public:

  /**
     Reads the file fname and sets all values.
     @param argc The number of arguments in argv
     @param argv All arguments supplied to the program.
  */
  OptionReader(int argc, char **argv)
    {
      argC=argc;
      argV=argv;

      noLink=false;

      mainfile=NULL;
      executable=NULL;
      linker=NULL;
      linkerTail=NULL;
      compiler=NULL;
      makescript=NULL;
      dotoutput=NULL;

      makFileFound=0;
      for(int i=1;i<argc;i++)
	{
	  if (argv[i][0]=='-')
	    {
	      parseLine(argv[i]+1);
	    }
	  else if (!makFileFound)
	    {
	      makFileFound=i;
	      FILE *fin=fopen(argv[i],"r");
	      if (fin)
		{
		  parse(fin);
		  fclose(fin);
		}
	      else
		{
		  printf("Failed to open file: %s\n",argv[i]);
		}
	    }
	}
      if (makFileFound>0)
	{
	  defaults(argv[makFileFound]);
	}
      else
	{
	  defaults(NULL);
	}
    }

  /**
     Deletes all strings.
  */
  ~OptionReader()
    {
      delete [] mainfile;
      delete [] executable;
      delete [] linker;
      delete [] linkerTail;
      delete [] compiler;
      delete [] makescript;
      delete [] dotoutput;
    }

  /**
     Gets the arguments that was passed to the OptionReader in the 
     constructor.
     @param argv Will be set to point at the arguments.
     @return The number of arguments in argv.
  */
  int getArguments(char **&argv)
  {
    argv=argV;
    return argC;
  }

  /**
     Dumps the content of this OptionReader object on stdout.
  */
  void print(void)
    {
      printf("Mainfile: %s\n",mainfile);
      printf("Executable: %s\n",executable);
      printf("Linker: %s <...> %s\n",linker,linkerTail);
      printf("Compiler: %s\n",compiler);
      printf("Makescript: %s\n",makescript);
      printf("Dotoutput: %s\n",dotoutput);
    }

  /**
     @return The mainfile complete path.
  */
  char * getMainfile(void)
    {
      return mainfile;
    }

  /**
     @return The executable complete path.
  */
  char * getExecutable(void)
    {
      return executable;
    }

  /**
     @return The link flags.
  */
  char * getLinker(void)
    {
      return linker;
    }

  /**
     @return The link tail flags.
  */
  char * getLinkerTail(void)
    {
      return linkerTail;
    }

  /**
     @return The compiler complete path and flags.
  */
  char * getCompiler(void)
    {
      return compiler;
    }

  /**
     @return The makescript complete path.
  */
  char * getMakescript(void)
    {
      return makescript;
    }

  /**
     @return The dotoutput complete path.
  */
  char * getDotoutput(void)
    {
      return dotoutput;
    }

  /**
     @return The index in the arguments for the options file.
  */
  int getOptionFileIndex(void)
  {
    return makFileFound;
  }

  /**
     @return True if the NOLINK option was specified.
  */
  bool performLink(void)
  {
    return !noLink;
  }

  /**
     Displays a short help.
  */
  void displayHelp(void)
  {
    printf("Command line: [opt] [optfile] [opt] [file1 file2 file3 ...] "
	   "[opt]\n\n"
	   
	   "fileX are all .h/.hh/.hpp files the mainfile could"
	   " possibly be depending on.\n\n"
	   
	   "Options format (preceeded by '-' when used as parameters):\n\n"
	   
	   "\tMAINFILE=somefile.cpp\n"
	   "\tEXECUTABLE=somefile\n"
	   "\tLINKER=g++ -lm -lxx -lxx -lxx -L. -L/usr/lib\n"
	   "\tLINKER_TAIL=libRPClort.a\n"
	   "\tCOMPILER=g++ -O -g\n"
	   "\tMAKESCRIPT=somefile.sh\n"
	   "\tDOTOUTPUT=somefile.dot\n"
	   "\tMAKER_ADD_DEPENDENCY=\"somefile.*\"\n"
	   "\t\tPlease note that the double quotes (\") are necessary!)\n"
	   "\tNOFILE\n\n");
  }

private:
  /**
     Parses the file lilne by line and extracts the variables.
     @param fin	A stream to read the optionfile from.
  */
  void parse(FILE *fin)
    {
      char buf[200000];
      if (fin)
	{
	  while (fgets(buf,199999,fin) && makFileFound>=0)
	    {
	      parseLine(buf);
	    }
	}
    }

  /**
     Parses the file lilne by line and extracts the variables.
     @param buf Any string from a optionsfile or parameter.
  */
  void parseLine(char *buf)
  {    
    if (strlen(buf) && buf[0]!='#' && buf[0]!='/')
      {
	char *p=strchr(buf,'=');
	if (!p)
	  {
	    if (buf[0])
	      {
		p=buf+strlen(buf)-1;
	      }
	  }
	if (p)
	  {
	    if (*p=='=')
	      {
		*p=0;
	      }
	    int length=strlen(p+1);
	    char **target=NULL;
	    while (p[length]=='\n' || p[length]=='\r')
	      {
		/* Strip linefeed */
		p[length]=0; 
		length--;
		if (length<0)
		  {
		    break;
		  }
	      }
	    if (!strcmp("MAINFILE",buf))
	      {
		target=&mainfile;
	      }
	    else if (!strcmp("EXECUTABLE",buf))
	      {
		target=&executable;
	      }
	    else if (!strcmp("LINKER",buf))
	      {
		target=&linker;
	      }
	    else if (!strcmp("LINKER_TAIL",buf))
	      {
		target=&linkerTail;
	      }
	    else if (!strcmp("COMPILER",buf))
	      {
		target=&compiler;
	      }
	    else if (!strcmp("MAKESCRIPT",buf))
	      {
		target=&makescript;
	      }
	    else if (!strcmp("DOTOUTPUT",buf))
	      {
		target=&dotoutput;
	      }
	    else if (!strcmp("MAKER_ADD_DEPENDENCY",buf))
	      {
		/* just ignore this token */
		target=NULL;
	      }
	    else if (!strcmp("NOFILE",buf))
	      {
		target=NULL;
		makFileFound=-1;
	      }
	    else if (!strcmp("NOLINK",buf))
	      {
		target=NULL;
		noLink=true;
	      }
	    else
	      {
		printf("Unknown token: %s\n",buf);
	      }
	    if (target) 
	      {
		*target=expandMacros(p+1);
	      }
	  }
	else
	  {
	    printf("This line seems invalid: %s",buf);
	  }
      }
  }

  /**
     Sets default values in all empty fields.
     @param fname The makescript file name.
  */
  void defaults(const char* fname)
    {
      if (!mainfile)
	{
	  if (fname)
	    {
	      mainfile=MakeUtilities::changeExtension(fname,"cpp");
	    }
	  else
	    {
	      printf("There must be a main file!\n");
	      _exit(-1);
	    }
	}
      if (!executable)
	{
	  char *p=MakeUtilities::strchrReverse(mainfile,'.');
	  if (!p)
	    {
	      printf("There must be a dot ('.') in mainfile.\n");
	      _exit(-1);
	    }
	  executable=MakeUtilities::getStrCopy(mainfile);
	  executable[p-mainfile]=0;
	}
      if (!linker)
	{
	  linker=MakeUtilities::getStrCopy("g++ -lm");
	}
      if (!linkerTail)
	{
	  linkerTail=MakeUtilities::getStrCopy("");
	}
      if (!compiler)
	{
	  compiler=MakeUtilities::getStrCopy("g++ -O2 -g");
	}
      if (!makescript)
	{
	  char *p=MakeUtilities::strchrReverse(mainfile,'.');
	  if (!p)
	    {
	      printf("There must be a dot ('.') in mainfile.\n");
	      _exit(-1);
	    }
	  makescript=MakeUtilities::getStrCopy(mainfile,3);
	  makescript[p-mainfile+1]=0;
	  strcat(makescript,"sh");
	}
    }

  /**
     Expands in input:
     $DATE$ to yyyy-mm-dd<br>
     $TIME$ to hh:mm
     @param input Any string.
     @result A pointer to a new string. CALLER MUST DELETE!
  */
  char *expandMacros(const char *input)
  {
    if (!input)
      {
	return NULL;
      }
    char *result=MakeUtilities::getStrCopy(input);

    char *p,*tmp;
    int len=strlen(result)+1;

    time_t now_t=time(NULL);
    struct tm *now=localtime(&now_t);

    p=strstr(result,"$DATE$");
    while(p)
      {
	*p=0;
	len+=4;
	tmp=new char[len];
	strcpy(tmp,result);
	sprintf(tmp-(result-p),"%4d-%.2d-%.2d%s",
		1900+now->tm_year,now->tm_mon+1,now->tm_mday,
		p+6);
	delete [] result;
	result=tmp;
	p=strstr(result,"$DATE$");
      }

    p=strstr(result,"$TIME$");
    while(p)
      {
	len--;
	sprintf(p,"%.2d:%.2d",
		now->tm_hour,now->tm_min);
	memmove(p+5,p+6,len-(p+6-result)+1);
	p=strstr(result,"$TIME$");
      }

    return result;
  }

  int makFileFound;
  char *mainfile, *executable, *linker, *linkerTail, *compiler, 
    *makescript, *dotoutput;
  char **argV;
  int argC;
  bool noLink;  
};

#endif

/*************** LOG *****************
 $Log: optionreader.h,v $
 Revision 1.10  2001/05/09 14:52:14  pb
 Added NOLINK option.

 Revision 1.9  2001/05/09 12:10:56  pb
 Cleaned up the source. Added new directive MAKER_ADD_DEPENDENCY.
 Added documentation.

 Revision 1.8  2001/05/08 18:10:18  pb
 Fixed a couple of bugs.

 Revision 1.7  2001/05/08 16:44:28  pb
 Minor fixs.

 Revision 1.6  2001/05/08 14:21:28  pb
 Works 100% and lots of new features.

 Revision 1.5  2001/05/07 17:17:15  pb
 It really works!

 Revision 1.4  2001/04/28 10:11:19  pb
 A -perhaps- working version.
		
 Revision 1.3  2001/04/27 14:23:36  pb
 It seems to start working. But there is a bug in building dependencies.

 Revision 1.2  2001/04/27 13:49:28  pb
 Added headers and footers.

*************************************/
